@@ -785,7 +785,7 @@ Interface.prototype.handleCommand = function(cmd) {
     self._lastCommand = null;
     self.tryQuit();
 
-  } else if (/^r(un)?$/.test(cmd)) {
+  } else if (/^r(un)?/.test(cmd)) {
     self._lastCommand = null;
     if (self.child) {
       self.restartQuestion(function(yes) {
@@ -805,7 +805,7 @@ Interface.prototype.handleCommand = function(cmd) {
       self.trySpawn();
     }
 
-  } else if (/^help$/.test(cmd)) {
+  } else if (/^help/.test(cmd)) {
     console.log(helpMessage);
     term.prompt();
 
@@ -819,7 +819,7 @@ Interface.prototype.handleCommand = function(cmd) {
       term.prompt();
     });
 
-  } else if (/^info(\s+breakpoint)?$/.test(cmd)) {
+  } else if (/info +breakpoints/.test(cmd)) {
     if (!client) {
       self.printNotConnected();
       return;
@@ -867,7 +867,7 @@ Interface.prototype.handleCommand = function(cmd) {
       term.prompt();
     });
 
-  } else if (/^b(ack)?t(race)?$/.test(cmd)) {
+  } else if (/^backtrace/.test(cmd) || /^bt/.test(cmd)) {
     if (!client) {
       self.printNotConnected();
       return;
@@ -905,7 +905,7 @@ Interface.prototype.handleCommand = function(cmd) {
     self.printScripts(cmd.indexOf('full') > 0);
     term.prompt();
 
-  } else if (/^c(ontinue)?$/.test(cmd)) {
+  } else if (/^c(ontinue)?/.test(cmd)) {
     if (!client) {
       self.printNotConnected();
       return;
@@ -916,7 +916,7 @@ Interface.prototype.handleCommand = function(cmd) {
       self.resume();
     });
 
-  } else if (/^k(ill)?$/.test(cmd)) {
+  } else if (/^k(ill)?/.test(cmd)) {
     if (!client) {
       self.printNotConnected();
       return;
@@ -934,7 +934,7 @@ Interface.prototype.handleCommand = function(cmd) {
       self.term.prompt();
     }
 
-  } else if (/^n(ext)?$/.test(cmd)) {
+  } else if (/^next/.test(cmd) || /^n/.test(cmd)) {
     if (!client) {
       self.printNotConnected();
       return;
@@ -943,7 +943,7 @@ Interface.prototype.handleCommand = function(cmd) {
       // Wait for break point. (disable raw mode?)
     });
 
-  } else if (/^s(tep)?$/.test(cmd)) {
+  } else if (/^step/.test(cmd) || /^s/.test(cmd)) {
     if (!client) {
       self.printNotConnected();
       return;
@@ -952,7 +952,7 @@ Interface.prototype.handleCommand = function(cmd) {
       // Wait for break point. (disable raw mode?)
     });
 
-  } else if (/^p(rint)?$/.test(cmd)) {
+  } else if (/^print/.test(cmd) || /^p/.test(cmd)) {
     if (!client) {
       self.printNotConnected();
       return;
@@ -998,7 +998,7 @@ Interface.prototype.yesNoQuestion = function(prompt, cb) {
       cb(false);
     } else {
       console.log('Please answer y or n.');
-      self.restartQuestion(cb);
+      self.yesNoQuestion(prompt, cb);
     }
   });
 };
