@@ -56,9 +56,21 @@ assert.AssertionError.prototype.toString = function() {
     return [this.name + ':', this.message].join(' ');
   } else {
     return [this.name + ':',
-            JSON.stringify(this.expected),
+            JSON.stringify(this.expected, replacer),
             this.operator,
-            JSON.stringify(this.actual)].join(' ');
+            JSON.stringify(this.actual, replacer)].join(' ');
+  }
+  function replacer(key, value) {
+    if (value === undefined) {
+      return '' + value;
+    }
+    if (typeof value === 'number' && (isNaN(value) || !isFinite(value))) {
+      return value.toString();
+    }
+    if (typeof value === 'function' || value instanceof RegExp) {
+      return value.toString();
+    }
+    return value;
   }
 };
 
