@@ -609,7 +609,6 @@ OutgoingMessage.prototype.write = function(chunk, encoding) {
     if (typeof(chunk) === 'string') {
       len = Buffer.byteLength(chunk, encoding);
       var chunk = len.toString(16) + CRLF + chunk + CRLF;
-      debug('string chunk = ' + util.inspect(chunk));
       ret = this._send(chunk, encoding);
     } else {
       // buffer
@@ -648,10 +647,19 @@ OutgoingMessage.prototype.addTrailers = function(headers) {
 
 
 OutgoingMessage.prototype.end = function(data, encoding) {
+  if (this.finished) {
+    return false;
+  }
   if (!this._header) {
     this._implicitHeader();
   }
 
+  if (data && !this._hasBody) {
+    console.error('This type of response MUST NOT have a body. ' +
+                  'Ignoring data passed to end().');
+    data = false;
+  }
+
   var ret;
 
   var hot = this._headerSent === false &&
@@ -667,6 +675,7 @@ OutgoingMessage.prototype.end = function(data, encoding) {
     //   res.writeHead();
     //   res.end(blah);
     // HACKY.
+
     if (this.chunkedEncoding) {
       var l = Buffer.byteLength(data, encoding).toString(16);
       ret = this.connection.write(this._header + l + CRLF +
@@ -798,6 +807,7 @@ ServerResponse.prototype.writeHead = function(statusCode) {
     reasonPhrase = STATUS_CODES[statusCode] || 'unknown';
     headerIndex = 1;
   }
+  this.statusCode = statusCode;
 
   var obj = arguments[headerIndex];
 
@@ -1152,12 +1162,6 @@ Agent.prototype.appendMessage = function(options) {
   this.queue.push(req);
   req._queue = this.queue;
 
-  /*
-  req.on('finish', function () {
-    self._cycle();
-  });
-  */
-
   this._cycle();
 
   return req;
@@ -1339,7 +1343,10 @@ Agent.prototype._establishNewConnection = function() {
         debug('AGENT socket keep-alive');
       }
 
-      req.detachSocket(socket);
+      // The socket may already be detached and destroyed by an abort call
+      if (socket._httpMessage) {
+        req.detachSocket(socket);
+      }
 
       assert(!socket._httpMessage);
 
@@ -1374,8 +1381,6 @@ Agent.prototype._cycle = function() {
   var first = this.queue[0];
   if (!first) return;
 
-  var haveConnectingSocket = false;
-
   // First try to find an available socket.
   for (var i = 0; i < this.sockets.length; i++) {
     var socket = this.sockets[i];
@@ -1395,13 +1400,11 @@ Agent.prototype._cycle = function() {
       self._cycle(); // try to dispatch another
       return;
     }
-
-    if (socket._httpConnecting) haveConnectingSocket = true;
   }
 
   // If no sockets are connecting, and we have space for another we should
   // be starting a new connection to handle this request.
-  if (!haveConnectingSocket && this.sockets.length < this.maxSockets) {
+  if (this.sockets.length < this.maxSockets) {
     this._establishNewConnection();
   }
 
