@@ -25,6 +25,15 @@ var runInThisContext = Script.runInThisContext;
 var runInNewContext = Script.runInNewContext;
 var assert = require('assert').ok;
 
+
+// If obj.hasOwnProperty has been overridden, then calling
+// obj.hasOwnProperty(prop) will break.
+// See: https://github.com/joyent/node/issues/1707
+function hasOwnProperty(obj, prop) {
+  return Object.prototype.hasOwnProperty.call(obj, prop);
+}
+
+
 function Module(id, parent) {
   this.id = id;
   this.exports = {};
@@ -85,7 +94,7 @@ function statPath(path) {
 var packageCache = {};
 
 function readPackage(requestPath) {
-  if (packageCache.hasOwnProperty(requestPath)) {
+  if (hasOwnProperty(packageCache, requestPath)) {
     return packageCache[requestPath];
   }
 
