@@ -409,16 +409,10 @@ Socket.prototype._writeOut = function(data, encoding, fd, cb) {
       allocNewPool();
     }
 
-    if (!encoding || encoding == 'utf8' || encoding == 'utf-8') {
-      // default to utf8
-      bytesWritten = pool.write(data, 'utf8', pool.used);
-      charsWritten = Buffer._charsWritten;
-    } else {
-      bytesWritten = pool.write(data, encoding, pool.used);
-      charsWritten = bytesWritten;
-    }
+    bytesWritten = pool.write(data, encoding, pool.used);
+    charsWritten = Buffer._charsWritten;
 
-    if (encoding && data.length > 0) {
+    if (data.length > 0) {
       assert(bytesWritten > 0);
     }
 
@@ -717,7 +711,13 @@ Socket.prototype.connect = function() {
     // TCP
     require('dns').lookup(arguments[1], function(err, ip, addressType) {
       if (err) {
-        self.emit('error', err);
+        // net.createConnection() creates a net.Socket object and
+        // immediately calls net.Socket.connect() on it (that's us).
+        // There are no event listeners registered yet so defer the
+        // error event to the next tick.
+        process.nextTick(function() {
+          self.emit('error', err);
+        });
       } else {
         timers.active(self);
         self.type = addressType == 4 ? 'tcp4' : 'tcp6';
@@ -1047,7 +1047,7 @@ Server.prototype.listen = function() {
         if (err.errno == ENOENT) {
           self._doListen(path);
         } else {
-          throw r;
+          throw err;
         }
       } else {
         if (!r.isSocket()) {
