@@ -46,6 +46,14 @@ var path = require('path');
 var fs = require('fs');
 var rl = require('readline');
 
+// If obj.hasOwnProperty has been overridden, then calling
+// obj.hasOwnProperty(prop) will break.
+// See: https://github.com/joyent/node/issues/1707
+function hasOwnProperty(obj, prop) {
+  return Object.prototype.hasOwnProperty.call(obj, prop);
+}
+
+
 var context;
 
 var disableColors = true;
@@ -446,7 +454,7 @@ REPLServer.prototype.complete = function(line) {
       group.sort();
       for (var j = 0; j < group.length; j++) {
         c = group[j];
-        if (!uniq.hasOwnProperty(c)) {
+        if (!hasOwnProperty(uniq, c)) {
           completions.push(c);
           uniq[c] = true;
         }
