@@ -47,7 +47,7 @@
 # include <grp.h> /* getgrnam() */
 #endif
 
-#include <platform.h>
+#include "platform.h"
 #include <node_buffer.h>
 #include <node_io_watcher.h>
 #include <node_net.h>
@@ -1574,6 +1574,7 @@ typedef void (*extInit)(Handle<Object> exports);
 // DLOpen is node.dlopen(). Used to load 'module.node' dynamically shared
 // objects.
 Handle<Value> DLOpen(const v8::Arguments& args) {
+  node_module_struct compat_mod;
   HandleScope scope;
 
   if (args.Length() < 2) return Undefined();
@@ -1616,10 +1617,13 @@ Handle<Value> DLOpen(const v8::Arguments& args) {
   // Get the init() function from the dynamically shared object.
   node_module_struct *mod = static_cast<node_module_struct *>(dlsym(handle, symstr));
   free(symstr);
+  symstr = NULL;
+
   // Error out if not found.
   if (mod == NULL) {
     /* Start Compatibility hack: Remove once everyone is using NODE_MODULE macro */
-    node_module_struct compat_mod;
+    memset(&compat_mod, 0, sizeof compat_mod);
+
     mod = &compat_mod;
     mod->version = NODE_MODULE_VERSION;
 
