@@ -1,14 +1,43 @@
+// Copyright Joyent, Inc. and other Node contributors.
+//
+// Permission is hereby granted, free of charge, to any person obtaining a
+// copy of this software and associated documentation files (the
+// "Software"), to deal in the Software without restriction, including
+// without limitation the rights to use, copy, modify, merge, publish,
+// distribute, sublicense, and/or sell copies of the Software, and to permit
+// persons to whom the Software is furnished to do so, subject to the
+// following conditions:
+//
+// The above copyright notice and this permission notice shall be included
+// in all copies or substantial portions of the Software.
+//
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
+// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
+// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
+// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
+// USE OR OTHER DEALINGS IN THE SOFTWARE.
+
+
+#include <node.h>
 #include <node_buffer.h>
 
+#include <v8.h>
+
 #include <assert.h>
 #include <stdlib.h> // malloc, free
-#include <v8.h>
-
 #include <string.h> // memcpy
 
-#include <arpa/inet.h>  // htons, htonl
+#ifdef __MINGW32__
+# include "platform.h"
+# include <platform_win32_winsock.h> // htons, htonl
+#endif
+
+#ifdef __POSIX__
+# include <arpa/inet.h> // htons, htonl
+#endif
 
-#include <node.h>
 
 #define MIN(a,b) ((a) < (b) ? (a) : (b))
 
@@ -43,49 +72,6 @@ static Persistent<String> write_sym;
 Persistent<FunctionTemplate> Buffer::constructor_template;
 
 
-// Each javascript Buffer object is backed by a Blob object.
-// the Blob is just a C-level chunk of bytes.
-// It has a reference count.
-struct Blob_ {
-  unsigned int refs;
-  size_t length;
-  char *data;
-};
-typedef struct Blob_ Blob;
-
-
-static inline Blob * blob_new(size_t length) {
-  Blob * blob  = (Blob*) malloc(sizeof(Blob));
-  if (!blob) return NULL;
-
-  blob->data = (char*) malloc(length);
-  if (!blob->data) {
-    free(blob);
-    return NULL;
-  }
-
-  V8::AdjustAmountOfExternalAllocatedMemory(sizeof(Blob) + length);
-  blob->length = length;
-  blob->refs = 0;
-  return blob;
-}
-
-
-static inline void blob_ref(Blob *blob) {
-  blob->refs++;
-}
-
-
-static inline void blob_unref(Blob *blob) {
-  assert(blob->refs > 0);
-  if (--blob->refs == 0) {
-    //fprintf(stderr, "free %d bytes\n", blob->length);
-    V8::AdjustAmountOfExternalAllocatedMemory(-(sizeof(Blob) + blob->length));
-    free(blob->data);
-    free(blob);
-  }
-}
-
 static inline size_t base64_decoded_size(const char *src, size_t size) {
   const char *const end = src + size;
   const int remainder = size % 4;
@@ -110,162 +96,141 @@ static inline size_t base64_decoded_size(const char *src, size_t size) {
   return size;
 }
 
-#if 0
-// When someone calls buffer.asciiSlice, data is not copied. Instead V8
-// references in the underlying Blob with this ExternalAsciiStringResource.
-class AsciiSliceExt: public String::ExternalAsciiStringResource {
- friend class Buffer;
- public:
-  AsciiSliceExt(Buffer *parent, size_t start, size_t end) {
-    blob_ = parent->blob();
-    blob_ref(blob_);
 
-    assert(start <= end);
-    length_ = end - start;
-    assert(start + length_ <= parent->length());
-    data_ = parent->data() + start;
-  }
-
-
-  ~AsciiSliceExt() {
-    //fprintf(stderr, "free ascii slice (%d refs left)\n", blob_->refs);
-    blob_unref(blob_);
-  }
-
-
-  const char* data() const { return data_; }
-  size_t length() const { return length_; }
-
- private:
-  const char *data_;
-  size_t length_;
-  Blob *blob_;
-};
-#endif
-
-Buffer* Buffer::New(size_t size) {
+static size_t ByteLength (Handle<String> string, enum encoding enc) {
   HandleScope scope;
 
-  Local<Value> arg = Integer::NewFromUnsigned(size);
+  if (enc == UTF8) {
+    return string->Utf8Length();
+  } else if (enc == BASE64) {
+    String::Utf8Value v(string);
+    return base64_decoded_size(*v, v.length());
+  } else if (enc == UCS2) {
+    return string->Length() * 2;
+  } else {
+    return string->Length();
+  }
+}
+
+
+Handle<Object> Buffer::New(Handle<String> string) {
+  HandleScope scope;
+
+  // get Buffer from global scope.
+  Local<Object> global = v8::Context::GetCurrent()->Global();
+  Local<Value> bv = global->Get(String::NewSymbol("Buffer"));
+  assert(bv->IsFunction());
+  Local<Function> b = Local<Function>::Cast(bv);
+
+  Local<Value> argv[1] = { Local<Value>::New(string) };
+  Local<Object> instance = b->NewInstance(1, argv);
+
+  return scope.Close(instance);
+}
+
+
+Buffer* Buffer::New(size_t length) {
+  HandleScope scope;
+
+  Local<Value> arg = Integer::NewFromUnsigned(length);
   Local<Object> b = constructor_template->GetFunction()->NewInstance(1, &arg);
+  if (b.IsEmpty()) return NULL;
 
   return ObjectWrap::Unwrap<Buffer>(b);
 }
 
 
-Handle<Value> Buffer::New(const Arguments &args) {
+Buffer* Buffer::New(char* data, size_t length) {
   HandleScope scope;
 
+  Local<Value> arg = Integer::NewFromUnsigned(0);
+  Local<Object> obj = constructor_template->GetFunction()->NewInstance(1, &arg);
+
+  Buffer *buffer = ObjectWrap::Unwrap<Buffer>(obj);
+  buffer->Replace(data, length, NULL, NULL);
+
+  return buffer;
+}
+
+
+Buffer* Buffer::New(char *data, size_t length,
+                    free_callback callback, void *hint) {
+  HandleScope scope;
+
+  Local<Value> arg = Integer::NewFromUnsigned(0);
+  Local<Object> obj = constructor_template->GetFunction()->NewInstance(1, &arg);
+
+  Buffer *buffer = ObjectWrap::Unwrap<Buffer>(obj);
+  buffer->Replace(data, length, callback, hint);
+
+  return buffer;
+}
+
+
+Handle<Value> Buffer::New(const Arguments &args) {
   if (!args.IsConstructCall()) {
-    Local<Value> argv[10];
-    for (int i = 0; i < MIN(args.Length(), 10); i++) {
-      argv[i] = args[i];
-    }
-    Local<Object> instance =
-      constructor_template->GetFunction()->NewInstance(args.Length(), argv);
-    return scope.Close(instance);
+    return FromConstructorTemplate(constructor_template, args);
   }
 
+  HandleScope scope;
+
   Buffer *buffer;
   if (args[0]->IsInt32()) {
     // var buffer = new Buffer(1024);
     size_t length = args[0]->Uint32Value();
-    buffer = new Buffer(length);
-
-  } else if (args[0]->IsArray()) {
-    Local<Array> a = Local<Array>::Cast(args[0]);
-    buffer = new Buffer(a->Length());
-    char *p = buffer->data();
-    for (int i = 0; i < a->Length(); i++) {
-      p[i] = a->Get(i)->Uint32Value();
-    }
-  } else if (args[0]->IsString()) {
-    Local<String> s = args[0]->ToString();
-    enum encoding e = ParseEncoding(args[1], UTF8);
-    int length = e == UTF8 ? s->Utf8Length() : s->Length();
-
-    // input gets base64-decoded, adjust buffer size
-    if (e == BASE64) {
-      const String::AsciiValue data(s);
-      length = base64_decoded_size(*data, data.length());
-    }
-
-    buffer = new Buffer(length);
-  } else if (Buffer::HasInstance(args[0]) && args.Length() > 2) {
-    // var slice = new Buffer(buffer, 123, 130);
-    // args: parent, start, end
-    Buffer *parent = ObjectWrap::Unwrap<Buffer>(args[0]->ToObject());
-    SLICE_ARGS(args[1], args[2])
-    buffer = new Buffer(parent, start, end);
+    buffer = new Buffer(args.This(), length);
   } else {
     return ThrowException(Exception::TypeError(String::New("Bad argument")));
   }
-
-  buffer->Wrap(args.This());
-  args.This()->SetIndexedPropertiesToExternalArrayData(buffer->data(),
-                                                       kExternalUnsignedByteArray,
-                                                       buffer->length());
-  args.This()->Set(length_symbol, Integer::New(buffer->length_));
-
-  if (args[0]->IsString()) {
-    if (write_sym.IsEmpty()) {
-      write_sym = Persistent<String>::New(String::NewSymbol("write"));
-    }
-
-    Local<Value> write_v = args.This()->Get(write_sym);
-    assert(write_v->IsFunction());
-    Local<Function> write = Local<Function>::Cast(write_v);
-
-    Local<Value> argv[2] = { args[0], args[1] };
-
-    TryCatch try_catch;
-
-    write->Call(args.This(), 2, argv);
-
-    if (try_catch.HasCaught()) {
-      FatalException(try_catch);
-    }
-  }
-
   return args.This();
 }
 
 
-Buffer::Buffer(size_t length) : ObjectWrap() {
-  blob_ = blob_new(length);
-  off_ = 0;
-  length_ = length;
+Buffer::Buffer(Handle<Object> wrapper, size_t length) : ObjectWrap() {
+  Wrap(wrapper);
 
-  blob_ref(blob_);
+  length_ = 0;
+  callback_ = NULL;
 
-  V8::AdjustAmountOfExternalAllocatedMemory(sizeof(Buffer));
-}
-
-
-Buffer::Buffer(Buffer *parent, size_t start, size_t end) : ObjectWrap() {
-  blob_ = parent->blob_;
-  assert(blob_->refs > 0);
-  blob_ref(blob_);
-
-  assert(start <= end);
-  off_ = parent->off_ + start;
-  length_ = end - start;
-  assert(length_ <= parent->length_);
-
-  V8::AdjustAmountOfExternalAllocatedMemory(sizeof(Buffer));
+  Replace(NULL, length, NULL, NULL);
 }
 
 
 Buffer::~Buffer() {
-  assert(blob_->refs > 0);
-  //fprintf(stderr, "free buffer (%d refs left)\n", blob_->refs);
-  blob_unref(blob_);
-  V8::AdjustAmountOfExternalAllocatedMemory(-static_cast<long int>(sizeof(Buffer)));
+  Replace(NULL, 0, NULL, NULL);
 }
 
 
-char* Buffer::data() {
-  return blob_->data + off_;
+void Buffer::Replace(char *data, size_t length,
+                     free_callback callback, void *hint) {
+  HandleScope scope;
+
+  if (callback_) {
+    callback_(data_, callback_hint_);
+  } else if (length_) {
+    delete [] data_;
+    V8::AdjustAmountOfExternalAllocatedMemory(-(sizeof(Buffer) + length_));
+  }
+
+  length_ = length;
+  callback_ = callback;
+  callback_hint_ = hint;
+
+  if (callback_) {
+    data_ = data;
+  } else if (length_) {
+    data_ = new char[length_];
+    if (data)
+      memcpy(data_, data, length_);
+    V8::AdjustAmountOfExternalAllocatedMemory(sizeof(Buffer) + length_);
+  } else {
+    data_ = NULL;
+  }
+
+  handle_->SetIndexedPropertiesToExternalArrayData(data_,
+                                                   kExternalUnsignedByteArray,
+                                                   length_);
+  handle_->Set(length_symbol, Integer::NewFromUnsigned(length_));
 }
 
 
@@ -274,7 +239,7 @@ Handle<Value> Buffer::BinarySlice(const Arguments &args) {
   Buffer *parent = ObjectWrap::Unwrap<Buffer>(args.This());
   SLICE_ARGS(args[0], args[1])
 
-  const char *data = const_cast<char*>(parent->data() + start);
+  char *data = parent->data_ + start;
   //Local<String> string = String::New(data, end - start);
 
   Local<Value> b =  Encode(data, end - start, BINARY);
@@ -288,18 +253,9 @@ Handle<Value> Buffer::AsciiSlice(const Arguments &args) {
   Buffer *parent = ObjectWrap::Unwrap<Buffer>(args.This());
   SLICE_ARGS(args[0], args[1])
 
-#if 0
-  AsciiSliceExt *ext = new AsciiSliceExt(parent, start, end);
-  Local<String> string = String::NewExternal(ext);
-  // There should be at least two references to the blob now - the parent
-  // and the slice.
-  assert(parent->blob_->refs >= 2);
-#endif
-
-  const char *data = const_cast<char*>(parent->data() + start);
+  char* data = parent->data_ + start;
   Local<String> string = String::New(data, end - start);
 
-
   return scope.Close(string);
 }
 
@@ -308,24 +264,42 @@ Handle<Value> Buffer::Utf8Slice(const Arguments &args) {
   HandleScope scope;
   Buffer *parent = ObjectWrap::Unwrap<Buffer>(args.This());
   SLICE_ARGS(args[0], args[1])
-  const char *data = const_cast<char*>(parent->data() + start);
+  char *data = parent->data_ + start;
   Local<String> string = String::New(data, end - start);
   return scope.Close(string);
 }
 
+Handle<Value> Buffer::Ucs2Slice(const Arguments &args) {
+  HandleScope scope;
+  Buffer *parent = ObjectWrap::Unwrap<Buffer>(args.This());
+  SLICE_ARGS(args[0], args[1])
+  uint16_t *data = (uint16_t*)(parent->data_ + start);
+  Local<String> string = String::New(data, (end - start) / 2);
+  return scope.Close(string);
+}
+
 static const char *base64_table = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                                   "abcdefghijklmnopqrstuvwxyz"
                                   "0123456789+/";
 static const int unbase64_table[] =
-  {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
+  {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-2,-1,-1,-2,-1,-1
   ,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
-  ,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,62,-1,-1,-1,63
+  ,-2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,62,-1,-1,-1,63
   ,52,53,54,55,56,57,58,59,60,61,-1,-1,-1,-1,-1,-1
   ,-1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14
   ,15,16,17,18,19,20,21,22,23,24,25,-1,-1,-1,-1,-1
   ,-1,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40
   ,41,42,43,44,45,46,47,48,49,50,51,-1,-1,-1,-1,-1
+  ,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
+  ,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
+  ,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
+  ,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
+  ,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
+  ,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
+  ,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
+  ,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
   };
+#define unbase64(x) unbase64_table[(uint8_t)(x)]
 
 
 Handle<Value> Buffer::Base64Slice(const Arguments &args) {
@@ -344,10 +318,10 @@ Handle<Value> Buffer::Base64Slice(const Arguments &args) {
   bool b1_oob, b2_oob;
 
   while (i < end) {
-    bitbuf[0] = parent->data()[i++];
+    bitbuf[0] = parent->data_[i++];
 
     if (i < end) {
-      bitbuf[1] = parent->data()[i];
+      bitbuf[1] = parent->data_[i];
       b1_oob = false;
     }  else {
       bitbuf[1] = 0;
@@ -356,7 +330,7 @@ Handle<Value> Buffer::Base64Slice(const Arguments &args) {
     i++;
 
     if (i < end) {
-      bitbuf[2] = parent->data()[i];
+      bitbuf[2] = parent->data_[i];
       b2_oob = false;
     }  else {
       bitbuf[2] = 0;
@@ -367,12 +341,12 @@ Handle<Value> Buffer::Base64Slice(const Arguments &args) {
 
     c = bitbuf[0] >> 2;
     assert(c < 64);
-    out[j++] = base64_table[c];
+    out[j++] = base64_table[(int)c];
     assert(j < out_len);
 
     c = ((bitbuf[0] & 0x03) << 4) | (bitbuf[1] >> 4);
     assert(c < 64);
-    out[j++] = base64_table[c];
+    out[j++] = base64_table[(int)c];
     assert(j < out_len);
 
     if (b1_oob) {
@@ -380,7 +354,7 @@ Handle<Value> Buffer::Base64Slice(const Arguments &args) {
     } else {
       c = ((bitbuf[1] & 0x0F) << 2) | (bitbuf[2] >> 6);
       assert(c < 64);
-      out[j++] = base64_table[c];
+      out[j++] = base64_table[(int)c];
     }
     assert(j < out_len);
 
@@ -389,7 +363,7 @@ Handle<Value> Buffer::Base64Slice(const Arguments &args) {
     } else {
       c = bitbuf[2] & 0x3F;
       assert(c < 64);
-      out[j++]  = base64_table[c];
+      out[j++]  = base64_table[(int)c];
     }
     assert(j <= out_len);
   }
@@ -400,15 +374,6 @@ Handle<Value> Buffer::Base64Slice(const Arguments &args) {
 }
 
 
-Handle<Value> Buffer::Slice(const Arguments &args) {
-  HandleScope scope;
-  Local<Value> argv[3] = { args.This(), args[0], args[1] };
-  Local<Object> slice =
-    constructor_template->GetFunction()->NewInstance(3, argv);
-  return scope.Close(slice);
-}
-
-
 // var bytesCopied = buffer.copy(target, targetStart, sourceStart, sourceEnd);
 Handle<Value> Buffer::Copy(const Arguments &args) {
   HandleScope scope;
@@ -420,12 +385,14 @@ Handle<Value> Buffer::Copy(const Arguments &args) {
             "First arg should be a Buffer")));
   }
 
-  Buffer *target = ObjectWrap::Unwrap<Buffer>(args[0]->ToObject());
+  Local<Object> target = args[0]->ToObject();
+  char *target_data = Buffer::Data(target);
+  ssize_t target_length = Buffer::Length(target);
 
   ssize_t target_start = args[1]->Int32Value();
   ssize_t source_start = args[2]->Int32Value();
   ssize_t source_end = args[3]->IsInt32() ? args[3]->Int32Value()
-                                          : source->length();
+                                          : source->length_;
 
   if (source_end < source_start) {
     return ThrowException(Exception::Error(String::New(
@@ -437,40 +404,36 @@ Handle<Value> Buffer::Copy(const Arguments &args) {
     return scope.Close(Integer::New(0));
   }
 
-  if (target_start < 0 || target_start >= target->length()) {
+  if (target_start < 0 || target_start >= target_length) {
     return ThrowException(Exception::Error(String::New(
             "targetStart out of bounds")));
   }
 
-  if (source_start < 0 || source_start >= source->length()) {
+  if (source_start < 0 || source_start >= source->length_) {
     return ThrowException(Exception::Error(String::New(
             "sourceStart out of bounds")));
   }
 
-  if (source_end < 0 || source_end > source->length()) {
+  if (source_end < 0 || source_end > source->length_) {
     return ThrowException(Exception::Error(String::New(
             "sourceEnd out of bounds")));
   }
 
-  ssize_t to_copy = MIN(source_end - source_start,
-                        target->length() - target_start);
+  ssize_t to_copy = MIN(MIN(source_end - source_start,
+                            target_length - target_start),
+                            source->length_ - source_start);
 
-  if (target->blob_ == source->blob_) {
-    // need to use slightly slower memmove is the ranges might overlap
-    memmove((void*)(target->data() + target_start),
-            (const void*)(source->data() + source_start),
-            to_copy);
-  } else {
-    memcpy((void*)(target->data() + target_start),
-           (const void*)(source->data() + source_start),
-           to_copy);
-  }
+
+  // need to use slightly slower memmove is the ranges might overlap
+  memmove((void *)(target_data + target_start),
+          (const void*)(source->data_ + source_start),
+          to_copy);
 
   return scope.Close(Integer::New(to_copy));
 }
 
 
-// var charsWritten = buffer.utf8Write(string, offset);
+// var charsWritten = buffer.utf8Write(string, offset, [maxLength]);
 Handle<Value> Buffer::Utf8Write(const Arguments &args) {
   HandleScope scope;
   Buffer *buffer = ObjectWrap::Unwrap<Buffer>(args.This());
@@ -482,31 +445,86 @@ Handle<Value> Buffer::Utf8Write(const Arguments &args) {
 
   Local<String> s = args[0]->ToString();
 
-  size_t offset = args[1]->Int32Value();
+  size_t offset = args[1]->Uint32Value();
 
-  if (s->Utf8Length() > 0 && offset >= buffer->length_) {
+  int length = s->Length();
+
+  if (length == 0) {
+    constructor_template->GetFunction()->Set(chars_written_sym,
+                                             Integer::New(0));
+    return scope.Close(Integer::New(0));
+  }
+
+  if (length > 0 && offset >= buffer->length_) {
     return ThrowException(Exception::TypeError(String::New(
             "Offset is out of bounds")));
   }
 
-  const char *p = buffer->data() + offset;
+  size_t max_length = args[2]->IsUndefined() ? buffer->length_ - offset
+                                             : args[2]->Uint32Value();
+  max_length = MIN(buffer->length_ - offset, max_length);
+
+  char* p = buffer->data_ + offset;
 
   int char_written;
 
-  int written = s->WriteUtf8((char*)p,
-                             buffer->length_ - offset,
+  int written = s->WriteUtf8(p,
+                             max_length,
                              &char_written,
                              String::HINT_MANY_WRITES_EXPECTED);
 
   constructor_template->GetFunction()->Set(chars_written_sym,
                                            Integer::New(char_written));
 
-  if (written > 0 && p[written-1] == '\0') written--;
+  if (written > 0 && p[written-1] == '\0' && char_written == length) {
+    uint16_t last_char;
+    s->Write(&last_char, length - 1, 1, String::NO_HINTS);
+    if (last_char != 0 || written > s->Utf8Length()) {
+      written--;
+    }
+  }
 
   return scope.Close(Integer::New(written));
 }
 
 
+// var charsWritten = buffer.ucs2Write(string, offset, [maxLength]);
+Handle<Value> Buffer::Ucs2Write(const Arguments &args) {
+  HandleScope scope;
+  Buffer *buffer = ObjectWrap::Unwrap<Buffer>(args.This());
+
+  if (!args[0]->IsString()) {
+    return ThrowException(Exception::TypeError(String::New(
+            "Argument must be a string")));
+  }
+
+  Local<String> s = args[0]->ToString();
+
+  size_t offset = args[1]->Uint32Value();
+
+  if (s->Length() > 0 && offset >= buffer->length_) {
+    return ThrowException(Exception::TypeError(String::New(
+            "Offset is out of bounds")));
+  }
+
+  size_t max_length = args[2]->IsUndefined() ? buffer->length_ - offset
+                                             : args[2]->Uint32Value();
+  max_length = MIN(buffer->length_ - offset, max_length) / 2;
+
+  uint16_t* p = (uint16_t*)(buffer->data_ + offset);
+
+  int written = s->Write(p,
+                         0,
+                         max_length,
+                         String::HINT_MANY_WRITES_EXPECTED);
+
+  constructor_template->GetFunction()->Set(chars_written_sym,
+                                           Integer::New(written));
+
+  return scope.Close(Integer::New(written * 2));
+}
+
+
 // var charsWritten = buffer.asciiWrite(string, offset);
 Handle<Value> Buffer::AsciiWrite(const Arguments &args) {
   HandleScope scope;
@@ -527,24 +545,38 @@ Handle<Value> Buffer::AsciiWrite(const Arguments &args) {
             "Offset is out of bounds")));
   }
 
-  const char *p = buffer->data() + offset;
+  size_t max_length = args[2]->IsUndefined() ? buffer->length_ - offset
+                                             : args[2]->Uint32Value();
+  max_length = MIN(s->Length(), MIN(buffer->length_ - offset, max_length));
 
-  size_t towrite = MIN((unsigned long) s->Length(), buffer->length_ - offset);
+  char *p = buffer->data_ + offset;
+
+  int written = s->WriteAscii(p,
+                              0,
+                              max_length,
+                              String::HINT_MANY_WRITES_EXPECTED);
+
+  constructor_template->GetFunction()->Set(chars_written_sym,
+                                           Integer::New(written));
 
-  int written = s->WriteAscii((char*)p, 0, towrite, String::HINT_MANY_WRITES_EXPECTED);
   return scope.Close(Integer::New(written));
 }
 
-// var bytesWritten = buffer.base64Write(string, offset);
+
+// var bytesWritten = buffer.base64Write(string, offset, [maxLength]);
 Handle<Value> Buffer::Base64Write(const Arguments &args) {
   HandleScope scope;
 
-  assert(unbase64_table['/'] == 63);
-  assert(unbase64_table['+'] == 62);
-  assert(unbase64_table['T'] == 19);
-  assert(unbase64_table['Z'] == 25);
-  assert(unbase64_table['t'] == 45);
-  assert(unbase64_table['z'] == 51);
+  assert(unbase64('/') == 63);
+  assert(unbase64('+') == 62);
+  assert(unbase64('T') == 19);
+  assert(unbase64('Z') == 25);
+  assert(unbase64('t') == 45);
+  assert(unbase64('z') == 51);
+
+  assert(unbase64(' ') == -2);
+  assert(unbase64('\n') == -2);
+  assert(unbase64('\r') == -2);
 
   Buffer *buffer = ObjectWrap::Unwrap<Buffer>(args.This());
 
@@ -574,29 +606,50 @@ Handle<Value> Buffer::Base64Write(const Arguments &args) {
   }
 
   char a, b, c, d;
-  char *dst = buffer->data();
+  char* start = buffer->data_ + offset;
+  char* dst = start;
   const char *src = *s;
   const char *const srcEnd = src + s.length();
 
   while (src < srcEnd) {
-    const int remaining = srcEnd - src;
-    if (remaining == 0 || *src == '=') break;
-    a = unbase64_table[*src++];
+    int remaining = srcEnd - src;
 
-    if (remaining == 1 || *src == '=') break;
-    b = unbase64_table[*src++];
+    while (unbase64(*src) < 0 && src < srcEnd) {
+      src++;
+      remaining--;
+    }
+    if (remaining == 0 || *src == '=') break;
+    a = unbase64(*src++);
+
+    while (unbase64(*src) < 0 && src < srcEnd) {
+      src++;
+      remaining--;
+    }
+    if (remaining <= 1 || *src == '=') break;
+    b = unbase64(*src++);
     *dst++ = (a << 2) | ((b & 0x30) >> 4);
 
-    if (remaining == 2 || *src == '=') break;
-    c = unbase64_table[*src++];
+    while (unbase64(*src) < 0 && src < srcEnd) {
+      src++;
+      remaining--;
+    }
+    if (remaining <= 2 || *src == '=') break;
+    c = unbase64(*src++);
     *dst++ = ((b & 0x0F) << 4) | ((c & 0x3C) >> 2);
 
-    if (remaining == 3 || *src == '=') break;
-    d = unbase64_table[*src++];
+    while (unbase64(*src) < 0 && src < srcEnd) {
+      src++;
+      remaining--;
+    }
+    if (remaining <= 3 || *src == '=') break;
+    d = unbase64(*src++);
     *dst++ = ((c & 0x03) << 6) | (d & 0x3F);
   }
 
-  return scope.Close(Integer::New(size));
+  constructor_template->GetFunction()->Set(chars_written_sym,
+                                           Integer::New(s.length()));
+
+  return scope.Close(Integer::New(dst - start));
 }
 
 
@@ -619,79 +672,21 @@ Handle<Value> Buffer::BinaryWrite(const Arguments &args) {
             "Offset is out of bounds")));
   }
 
-  char *p = (char*)buffer->data() + offset;
+  char *p = (char*)buffer->data_ + offset;
 
-  size_t towrite = MIN((unsigned long) s->Length(), buffer->length_ - offset);
+  size_t max_length = args[2]->IsUndefined() ? buffer->length_ - offset
+                                             : args[2]->Uint32Value();
+  max_length = MIN(s->Length(), MIN(buffer->length_ - offset, max_length));
+
+  int written = DecodeWrite(p, max_length, s, BINARY);
+
+  constructor_template->GetFunction()->Set(chars_written_sym,
+                                           Integer::New(written));
 
-  int written = DecodeWrite(p, towrite, s, BINARY);
   return scope.Close(Integer::New(written));
 }
 
 
-// buffer.unpack(format, index);
-// Starting at 'index', unpacks binary from the buffer into an array.
-// 'format' is a string
-//
-//  FORMAT  RETURNS
-//    N     uint32_t   a 32bit unsigned integer in network byte order
-//    n     uint16_t   a 16bit unsigned integer in network byte order
-//    o     uint8_t    a 8bit unsigned integer
-Handle<Value> Buffer::Unpack(const Arguments &args) {
-  HandleScope scope;
-  Buffer *buffer = ObjectWrap::Unwrap<Buffer>(args.This());
-
-  if (!args[0]->IsString()) {
-    return ThrowException(Exception::TypeError(String::New(
-            "Argument must be a string")));
-  }
-
-  String::AsciiValue format(args[0]->ToString());
-  uint32_t index = args[1]->Uint32Value();
-
-#define OUT_OF_BOUNDS ThrowException(Exception::Error(String::New("Out of bounds")))
-
-  Local<Array> array = Array::New(format.length());
-
-  uint8_t  uint8;
-  uint16_t uint16;
-  uint32_t uint32;
-
-  for (int i = 0; i < format.length(); i++) {
-    switch ((*format)[i]) {
-      // 32bit unsigned integer in network byte order
-      case 'N':
-        if (index + 3 >= buffer->length_) return OUT_OF_BOUNDS;
-        uint32 = htonl(*(uint32_t*)(buffer->data() + index));
-        array->Set(Integer::New(i), Integer::NewFromUnsigned(uint32));
-        index += 4;
-        break;
-
-      // 16bit unsigned integer in network byte order
-      case 'n':
-        if (index + 1 >= buffer->length_) return OUT_OF_BOUNDS;
-        uint16 = htons(*(uint16_t*)(buffer->data() + index));
-        array->Set(Integer::New(i), Integer::NewFromUnsigned(uint16));
-        index += 2;
-        break;
-
-      // a single octet, unsigned.
-      case 'o':
-        if (index >= buffer->length_) return OUT_OF_BOUNDS;
-        uint8 = (uint8_t)buffer->data()[index];
-        array->Set(Integer::New(i), Integer::NewFromUnsigned(uint8));
-        index += 1;
-        break;
-
-      default:
-        return ThrowException(Exception::Error(
-              String::New("Unknown format character")));
-    }
-  }
-
-  return scope.Close(array);
-}
-
-
 // var nbytes = Buffer.byteLength("string", "utf8")
 Handle<Value> Buffer::ByteLength(const Arguments &args) {
   HandleScope scope;
@@ -704,28 +699,45 @@ Handle<Value> Buffer::ByteLength(const Arguments &args) {
   Local<String> s = args[0]->ToString();
   enum encoding e = ParseEncoding(args[1], UTF8);
 
-  Local<Integer> length =
-    Integer::New(e == UTF8 ? s->Utf8Length() : s->Length());
-
-  return scope.Close(length);
+  return scope.Close(Integer::New(node::ByteLength(s, e)));
 }
 
 
 Handle<Value> Buffer::MakeFastBuffer(const Arguments &args) {
   HandleScope scope;
 
+  if (!Buffer::HasInstance(args[0])) {
+    return ThrowException(Exception::TypeError(String::New(
+            "First argument must be a Buffer")));
+  }
+
   Buffer *buffer = ObjectWrap::Unwrap<Buffer>(args[0]->ToObject());
   Local<Object> fast_buffer = args[1]->ToObject();;
   uint32_t offset = args[2]->Uint32Value();
   uint32_t length = args[3]->Uint32Value();
 
-  fast_buffer->SetIndexedPropertiesToPixelData((uint8_t*)buffer->data() + offset,
-                                               length);
+  fast_buffer->SetIndexedPropertiesToExternalArrayData(buffer->data_ + offset,
+                                                       kExternalUnsignedByteArray,
+                                                       length);
 
   return Undefined();
 }
 
 
+bool Buffer::HasInstance(v8::Handle<v8::Value> val) {
+  if (!val->IsObject()) return false;
+  v8::Local<v8::Object> obj = val->ToObject();
+
+  if (obj->GetIndexedPropertiesExternalArrayDataType() == kExternalUnsignedByteArray)
+    return true;
+
+  // Also check for SlowBuffers that are empty.
+  if (constructor_template->HasInstance(obj))
+    return true;
+
+  return false;
+}
+
 
 void Buffer::Initialize(Handle<Object> target) {
   HandleScope scope;
@@ -736,13 +748,13 @@ void Buffer::Initialize(Handle<Object> target) {
   Local<FunctionTemplate> t = FunctionTemplate::New(Buffer::New);
   constructor_template = Persistent<FunctionTemplate>::New(t);
   constructor_template->InstanceTemplate()->SetInternalFieldCount(1);
-  constructor_template->SetClassName(String::NewSymbol("Buffer"));
+  constructor_template->SetClassName(String::NewSymbol("SlowBuffer"));
 
   // copy free
   NODE_SET_PROTOTYPE_METHOD(constructor_template, "binarySlice", Buffer::BinarySlice);
   NODE_SET_PROTOTYPE_METHOD(constructor_template, "asciiSlice", Buffer::AsciiSlice);
   NODE_SET_PROTOTYPE_METHOD(constructor_template, "base64Slice", Buffer::Base64Slice);
-  NODE_SET_PROTOTYPE_METHOD(constructor_template, "slice", Buffer::Slice);
+  NODE_SET_PROTOTYPE_METHOD(constructor_template, "ucs2Slice", Buffer::Ucs2Slice);
   // TODO NODE_SET_PROTOTYPE_METHOD(t, "utf16Slice", Utf16Slice);
   // copy
   NODE_SET_PROTOTYPE_METHOD(constructor_template, "utf8Slice", Buffer::Utf8Slice);
@@ -751,7 +763,7 @@ void Buffer::Initialize(Handle<Object> target) {
   NODE_SET_PROTOTYPE_METHOD(constructor_template, "asciiWrite", Buffer::AsciiWrite);
   NODE_SET_PROTOTYPE_METHOD(constructor_template, "binaryWrite", Buffer::BinaryWrite);
   NODE_SET_PROTOTYPE_METHOD(constructor_template, "base64Write", Buffer::Base64Write);
-  NODE_SET_PROTOTYPE_METHOD(constructor_template, "unpack", Buffer::Unpack);
+  NODE_SET_PROTOTYPE_METHOD(constructor_template, "ucs2Write", Buffer::Ucs2Write);
   NODE_SET_PROTOTYPE_METHOD(constructor_template, "copy", Buffer::Copy);
 
   NODE_SET_METHOD(constructor_template->GetFunction(),
@@ -761,7 +773,7 @@ void Buffer::Initialize(Handle<Object> target) {
                   "makeFastBuffer",
                   Buffer::MakeFastBuffer);
 
-  target->Set(String::NewSymbol("Buffer"), constructor_template->GetFunction());
+  target->Set(String::NewSymbol("SlowBuffer"), constructor_template->GetFunction());
 }
 
 
